/*
 * HFS (aka MAC) filesystem/handler for MorphOS/AmigaOS
 * Author: Marek 'March' Szyprowski <marek@amiga.pl>
 */

#include <exec/types.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/filehandler.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/utility.h>

#include <string.h>
#include <ctype.h>

#include "mac_fs.h"
#include "mac_protos.h"

/* super block */

LONG ReadSuper (struct FSSuper *sb)
{
	struct DosEnvec *de = BADDR(glob->fssm->fssm_Environ);
	kprintf("\tTrying to open volume via libhfs\n");

	if ((sb->vol = hfs_mount((const char *)glob->fssm, 0, HFS_MODE_ANY)) != NULL)
	{
		hfsvolent volent;

		if (hfs_vstat(sb->vol, &volent) >= 0)
		{
			strncpy(&sb->name[1], volent.name, sizeof(sb->name) - 1);
			sb->name[sizeof(sb->name)-1] = 0;
			sb->name[0] = strlen(&sb->name[1]);

			sb->create_date = volent.crdate;

			sb->total_blocks = volent.totbytes >> HFS_BLOCKSZ_BITS;	  /* total bytes on volume */
			sb->free_blocks = volent.freebytes >> HFS_BLOCKSZ_BITS;	 /* free bytes on volume */

			sb->blocksize = HFS_BLOCKSZ;  /* volume allocation block size */

			kprintf("\tVolume successfully opened: %s\n", (LONG)volent.name);
			kprintf("\tTotal blocks: %ld\n", sb->total_blocks);
			kprintf("\tFree blocks: %ld\n", sb->free_blocks);
			kprintf("\tBlock size: %ld\n", volent.alblocksz);
			kprintf("\tCurrent blessed cnid: %ld\n", volent.blessed);

			if (1LL * sb->total_blocks * HFS_BLOCKSZ <= 1LL * de->de_Surfaces * de->de_BlocksPerTrack * (de->de_HighCyl+1 - de->de_LowCyl) / de->de_SectorPerBlock * (de->de_SizeBlock << 2))
			{
				return 0;
			}
			else
			{
				kprintf("\tError: too large filesystem for this partition!\n");
			}
		}
		hfs_umount(sb->vol);
		sb->vol = NULL;
	}
	else
	{
		kprintf("\tError: %s\n", (LONG)hfs_error);
	}
 
	kprintf("\tError opening volume\n");
	return ERROR_OBJECT_NOT_FOUND;
}

void FreeSuper(struct FSSuper *sb)
{
	kprintf("\tRemoving Super Block from memory\n");
	hfs_umount(sb->vol);
	sb->vol = NULL;
}

LONG CompareSuper(struct FSSuper *s1, struct FSSuper *s2)
{
	LONG res;

	if ((res = memcmp(s1->name, s2->name, s1->name[0])) != 0)
		return res;

	return s1->create_date - s2->create_date;
}

/* directory read */

LONG FillFIB (struct FSSuper *sb, ULONG cid, ULONG parid, struct FileInfoBlock *fib, BOOL dos64)
{
	hfsdirent ent;
	hfsdir *dir = NULL;
	ULONG size = 0;

	if (parid == HFS_CNID_ROOTPAR)
		dir = hfs_opendir(sb->vol, "");
	else
	{
		if (hfs_setcwd(sb->vol, parid) == 0)
			dir = hfs_opendir(sb->vol, ":");
	}

	if (dir == NULL)
	{
		kprintf("\terror - cannot open parent directory for cid %d\n", cid);
		return ERROR_OBJECT_NOT_FOUND;
	}

	while (hfs_readdir(dir, &ent) == 0)
		if (ent.cnid == cid)
			break;

	hfs_closedir(dir);
	if (ent.cnid != cid)
		return ERROR_OBJECT_NOT_FOUND;

	kprintf("\tFilling FIB data for key %ld\n", cid);
	if (ent.flags & HFS_ISDIR)
	{
		kprintf("\t\ttype: directory\n");

		if (cid != HFS_CNID_ROOTDIR)
			fib->fib_DirEntryType = ST_USERDIR;
		else
			fib->fib_DirEntryType = ST_ROOT;
	}
	else
	{
		kprintf("\t\ttype: file, creator %s, type %s\n", (ULONG)ent.u.file.creator, (ULONG)ent.u.file.type);
		fib->fib_DirEntryType = ST_FILE;
		size = ent.u.file.dsize;
	}

	fib->fib_DiskKey = cid;
	fib->fib_Size = size;
	fib->fib_NumBlocks = ((size + (HFS_BLOCKSZ - 1)) >> HFS_BLOCKSZ_BITS);

	if (dos64)
	{
		fib->fib_Size64 = size;
		fib->fib_NumBlocks64 = fib->fib_NumBlocks;
	}
	fib->fib_EntryType = fib->fib_DirEntryType;

	MacTime2DateStamp(ent.mddate, &fib->fib_Date);
	fib->fib_Comment[0] = '\0';
	strcpy(&fib->fib_FileName[1], ent.name);
	fib->fib_FileName[0] = strlen(ent.name);

	kprintf("\t\tname (len %ld) ", fib->fib_FileName[0]); knprints(&fib->fib_FileName[1], fib->fib_FileName[0]);
	fib->fib_Protection = 0;

	return 0;
}

LONG DirNextObj(struct FSSuper *sb, struct ExtFileLock *fl, struct FileInfoBlock *fib, BOOL dos64)
{
	ULONG cid = fl->fl_Key;
	LONG err = 0;
	BOOL rescan = FALSE;
	hfsdirent ent;

	kprintf("\tRead Next Object in directory %ld\n", cid);
	
	if (fib->fib_DiskKey == fl->fl_Key || fl->dir == NULL || fib->fib_DiskKey != fl->lastcid) /* (re)scan */
	{
		kprintf("\tSetting up initial directory state\n");

		if (fl->dir)
		{
			hfs_closedir(fl->dir);
			fl->dir = NULL;
		}

		if (cid == HFS_CNID_ROOTPAR)
			fl->dir = hfs_opendir(sb->vol, "");
		else
		{
			if (hfs_setcwd(sb->vol, cid) == 0)
				fl->dir = hfs_opendir(sb->vol, ":");
		}
 
		if (fl->dir == NULL)
		{
			kprintf("\tFailed to open directory\n");
			return ERROR_OBJECT_WRONG_TYPE;
		}
		rescan = TRUE;
	}

	if (fib->fib_DiskKey != fl->lastcid)
	{
		kprintf("\tLock cleared or altered between ExamineNext calls!\n");

		while (hfs_readdir(fl->dir, &ent) == 0)
			if (ent.cnid == fib->fib_DiskKey)
				break;

		if (ent.cnid != fib->fib_DiskKey)
			return ERROR_NO_MORE_ENTRIES;
	}
 
	if (hfs_readdir(fl->dir, &ent) == 0)
	{
		fl->lastcid = ent.cnid;
		err = FillFIB(sb, ent.cnid, ent.parid, fib, dos64);
	}
	else
		err = ERROR_NO_MORE_ENTRIES;

	return err;

}

LONG DirLookup(struct FSSuper *sb, ULONG dir_cid, UBYTE *name, ULONG namelen, ULONG *res_cid)
{
	UBYTE path[127];
	hfsdirent ent;

	if (namelen > sizeof(path))
		return ERROR_OBJECT_NOT_FOUND;

	memcpy(path, name, namelen);
	path[namelen] = '\0';

	if (hfs_setcwd(sb->vol, dir_cid) == 0)
	{
		if (hfs_stat(sb->vol, path, &ent) == 0)
		{
			*res_cid = ent.cnid;
			return 0;
		}
	}
	return ERROR_OBJECT_NOT_FOUND;
}

LONG GetParentDir(struct FSSuper *sb, ULONG cid, ULONG *res_parid)
{
	ULONG par_id = cid;
	if (hfs_dirinfo(sb->vol, &par_id, NULL) == 0)
	{
		*res_parid = par_id;
		return 0;
	}
	return ERROR_OBJECT_NOT_FOUND;
}

LONG FindObjectByPath(struct FSSuper *sb, struct ExtFileLock *fl, UBYTE *path, LONG pathlen, ULONG *dst_cid, ULONG *dst_parid)
{
	ULONG start_cid, start_parid;
	LONG err = 0;

	SkipColon(path, pathlen);
	kprintf("\tSearching for: "); knprints(path, pathlen);
 
	if (fl)
	{
		start_cid = fl->fl_Key;
		start_parid = fl->parid;
	}
	else
	{
		start_cid = HFS_CNID_ROOTDIR;
		start_parid = HFS_CNID_ROOTPAR;
	}

	while (err == 0)
	{
		ULONG len;
		kprintf("\tFindEntryByPath: start cid %ld pathlen %ld path ", start_cid, pathlen); knprints(path, pathlen);

		for (len=0; len < pathlen && path[len] != '/'; len++)
			;

		if (pathlen && *path == '/')
		{
			ULONG new_parid;
			kprintf("\tGetting parent directory.\n");
			if ((err = GetParentDir(sb, start_parid, &new_parid)) == 0)
			{
				start_cid = start_parid;
				start_parid = new_parid;
			}
		}
		else if (pathlen && (err = DirLookup(sb, start_cid, path, len, dst_cid)) == 0)
		{
			start_parid = start_cid;
			start_cid = *dst_cid;
		}
		kprintf("\tStart cid: %ld\n", start_cid);

		if (err == 0)
		{
			while (pathlen > 0)
			{
				pathlen--;
				if (*path++ == '/')
					break;
			}
			if (pathlen == 0)
				break;
		}
	}
	*dst_cid = start_cid;
	*dst_parid = start_parid;
 
	kprintf("\tFind by path result: cid %ld err %ld\n", *dst_cid, err);

	return err;
}

LONG LockByPath(struct FSSuper *sb, struct ExtFileLock *fl, UBYTE *name, LONG namelen, LONG access, BPTR *result)
{
	LONG err = ERROR_OBJECT_NOT_FOUND;
	ULONG cid, parid;

	if ((err = FindObjectByPath(sb, fl, name, namelen, &cid, &parid)) == 0)
		err = LockObject(sb, cid, parid, access, result);

	return err;
}

/* directory operations */

LONG CreateDirectory(struct FSSuper *sb, struct ExtFileLock *fl, UBYTE *path, LONG pathlen, BPTR *result)
{
	UBYTE *name;
	LONG namelen;
	LONG err = ERROR_OBJECT_NOT_FOUND;
	ULONG cid, parid;

	if (pathlen && path[pathlen-1] == '/')
		pathlen -= 1;

	kprintf("\tDirectory to be created: "); knprints(path, pathlen);
	SplitLastPart(path, pathlen, name, namelen);

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &cid, &parid)) == 0) /* find parent */
	{
		parid = cid;
		if ((err = DirLookup(sb, parid, name, namelen, &cid)) == ERROR_OBJECT_NOT_FOUND)
		{
			UBYTE dirname[128];
			memcpy(dirname, name, namelen);
			dirname[namelen] = '\0';
 
			kprintf("\tCreating directory in dir %ld: ", parid); knprints(name, namelen);

			if (hfs_setcwd(sb->vol, parid) == 0)
			{
				if (hfs_mkdir(sb->vol, dirname) == 0)
				{
					hfsdirent ent;
					hfs_flush(sb->vol);
					kprintf("\tDirectory created.\n");
					if (hfs_stat(sb->vol, dirname, &ent) == 0)
						err = LockObject(sb, ent.cnid, ent.parid, EXCLUSIVE_LOCK, result);
				}
			}
		}
		else if (err == 0)
		{
			kprintf("\tErr: Object exists\n");
			err = ERROR_OBJECT_EXISTS;
		}
	}
	return err;
}

LONG InternalDeleteObject(struct FSSuper *sb, ULONG parid, UBYTE *name, LONG namelen)
{
	LONG err = ERROR_OBJECT_NOT_FOUND;
	UBYTE objname[127];
	hfsdirent ent;

	if (namelen > sizeof(objname))
		return err;

	memcpy(objname, name, namelen);
	objname[namelen] = '\0';
 
	if (hfs_setcwd(sb->vol, parid) == 0)
	{
		if (hfs_stat(sb->vol, objname, &ent) == 0)
		{
			if ((err = IsLockable(sb, ent.cnid, EXCLUSIVE_LOCK)) == 0)
			{
				if (ent.flags & HFS_ISDIR)
				{
					if (hfs_rmdir(sb->vol, objname) == 0)
					{
						hfs_flush(sb->vol);
						kprintf("\tDirectory removed\n");
						err = 0;
					}
					else
						err = ERROR_DIRECTORY_NOT_EMPTY;
				}
				else
				{
					if (hfs_delete(sb->vol, objname) == 0)
					{
						hfs_flush(sb->vol);
						kprintf("\tFile deleted\n");
						err = 0;
					}
					else
						err = ERROR_OBJECT_WRONG_TYPE;
				}
			}
		}
	}
	return err;
}

LONG DeleteObject(struct FSSuper *sb, struct ExtFileLock *fl, UBYTE *path, LONG pathlen)
{
	UBYTE *name;
	LONG namelen;
	LONG err = ERROR_OBJECT_NOT_FOUND;
	ULONG cid, parid;

	if (pathlen && path[pathlen-1] == '/')
		pathlen -= 1;

	kprintf("\tDirectory to be created: "); knprints(path, pathlen);
	SplitLastPart(path, pathlen, name, namelen);

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &cid, &parid)) == 0) /* find parent */
	{
		parid = cid;
		err = InternalDeleteObject(sb, parid, name, namelen);
	}
	return err;
}

LONG RenameObject(struct FSSuper *sb, struct ExtFileLock *fl, UBYTE *path, LONG pathlen, struct ExtFileLock *dst_fl, UBYTE *dst_path, LONG dst_pathlen)
{
	ULONG src_cid, src_parid;
	LONG err;

	if (pathlen && path[pathlen-1] == '/')
		pathlen -= 1;

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &src_cid, &src_parid)) == 0) /* find source */
	{
		ULONG dst_cid, dst_parid;
		UBYTE *name, *dst_name;
		LONG namelen, dst_namelen;

		SplitLastPart(path, pathlen, name, namelen);
		SplitLastPart(dst_path, dst_pathlen, dst_name, dst_namelen);

		kprintf("\tFound old object in dir %ld, name: ", src_parid); knprints(name, namelen);

		if ((err = FindObjectByPath(sb, dst_fl, dst_path, dst_pathlen, &dst_cid, &dst_parid)) == 0) /* find destination parent */
		{
			ULONG cid = dst_cid;
			UBYTE *full_dst_path;
			UBYTE *p;
			LONG level=1; /* target object */

			kprintf("\tFound destination directory\n");

			while (hfs_dirinfo(sb->vol, &cid, NULL) == 0)
			{
				kprintf("\tDestination parent id: %ld\n", cid);
				level ++;
				if (cid == HFS_CNID_ROOTPAR)
					break;
			}
			if (cid != HFS_CNID_ROOTPAR)
				return ERROR_OBJECT_WRONG_TYPE;

			kprintf("\tDestination object level: %ld\n", level);
			if ((full_dst_path = FS_AllocMem(level * HFS_MAX_FLEN)) != NULL)
			{
				UBYTE tmpname[32];

				p = full_dst_path + (level * HFS_MAX_FLEN) - 1;
				*p = '\0';

				memcpy(p-dst_namelen, dst_name, dst_namelen);
				p -= dst_namelen;
				*--p = ':';

				cid = dst_cid;

				while (hfs_dirinfo(sb->vol, &cid, tmpname) == 0)
				{
					LONG tmpname_len = strlen(tmpname);
					memcpy(p-tmpname_len, tmpname, tmpname_len);
					p -= tmpname_len;
					*--p = ':';
 
					if (cid == HFS_CNID_ROOTPAR)
						break;
				}
				p++;
				kprintf("\tDestination object full path: %s\n", (LONG)p);

				err = ERROR_OBJECT_WRONG_TYPE;
				if (hfs_setcwd(sb->vol, src_parid) == 0)
				{
					UBYTE srcname[32];
					memcpy(srcname, name, namelen);
					srcname[namelen] = '\0';
					if ((hfs_rename(sb->vol, srcname, p)) == 0)
					{
						hfs_flush(sb->vol);
						kprintf("\tObject renamed\n");
						err = 0;
					}
					else
					{
						kprintf("\tError renaming object: %s\n", (LONG)hfs_error);
					}
				}
				FS_FreeMem(full_dst_path);
			}
			else
				err = ERROR_NO_FREE_STORE;
		}
	}
	return err;
}

/* files */

LONG FileOpen(struct FSSuper *sb, struct ExtFileLock *fl)
{
	ULONG cid = fl->fl_Key;
	if (hfs_setcwd(sb->vol, fl->parid) == 0)
	{
		hfsdir *dir;
		if ((dir = hfs_opendir(sb->vol, ":")) != NULL)
		{
			hfsdirent ent;

			while (hfs_readdir(dir, &ent) == 0)
				if (ent.cnid == cid)
					break;

			hfs_closedir(dir);
			if (ent.cnid == cid)
			{
				if ((fl->file = hfs_open(sb->vol, ent.name)) != NULL)
					return 0;
			}
		}
	}
	return ERROR_OBJECT_NOT_FOUND;
}

LONG CreateOpenFile(struct FSSuper *sb, struct ExtFileLock *fl, UBYTE *path, LONG pathlen, LONG action_type, struct FileHandle *fh)
{
	LONG err;
	ULONG cid, parid;
	LONG access = (action_type == ACTION_FINDOUTPUT) ? EXCLUSIVE_LOCK : SHARED_LOCK;
	BPTR lock = NULL;
	UBYTE *name;
	LONG namelen;

	kprintf("\tFile to be opened: "); knprints(path, pathlen);
	SplitLastPart(path, pathlen, name, namelen);

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &cid, &parid)) == 0) /* find parent */
	{
		parid = cid;
		if ((err = DirLookup(sb, parid, name, namelen, &cid)) == 0)
		{
			if (action_type == ACTION_FINDOUTPUT) /* need to delete existing object */
			{
				kprintf("\tFound object that need to be deleted %ld\n", cid);
				err = InternalDeleteObject(sb, parid, name, namelen);
			}
			else /* FINDINPUT or FINDUPDATE */
				err = LockObject(sb, cid, parid, access, &lock);
		}

		if ((err == 0 || err == ERROR_OBJECT_NOT_FOUND) && action_type != ACTION_FINDINPUT && lock == NULL)
		{
			hfsfile *newfile;
			UBYTE filename[128];
			memcpy(filename, name, namelen);
			filename[namelen] = '\0';

			/* create a file */
			kprintf("\tCreating file in dir %ld: ", parid); knprints(name, namelen);
			cid = 0;
			if ((newfile = hfs_create(sb->vol, filename, RAW_TYPE, RAW_CREA)) != NULL)
			{
				hfsdirent ent;
				hfs_flush(sb->vol);
				if (hfs_fstat(newfile, &ent) == 0)
				{
					cid = ent.cnid;
                    err = LockObject(sb, cid, parid, access, &lock);
				}
				else
					err = ERROR_OBJECT_WRONG_TYPE;
				hfs_close(newfile);
			}
			else
				err = ERROR_SEEK_ERROR;
		}

		if (err == 0 && lock)
		{
			struct ExtFileLock *fl_new = BADDR(lock);
			if ((err = FileOpen(glob->sb, fl_new)) == 0)
			{
				fh->fh_Arg1 = (LONG)lock;
				fh->fh_Port = DOSFALSE;
			}
			else
			{
				err = ERROR_OBJECT_WRONG_TYPE;
				FreeLock(fl_new);
			}
		}
	}
	return err;
}

LONG FileRead(struct FSSuper *sb, struct ExtFileLock *fl, LONG bytes, APTR buffer, LONG *res)
{
	if (fl->file == NULL)
		return ERROR_OBJECT_WRONG_TYPE;

	if ((*res = hfs_read(fl->file, buffer, bytes)) >= 0)
		return 0;
	return ERROR_SEEK_ERROR;
}

LONG FileWrite(struct FSSuper *sb, struct ExtFileLock *fl, LONG bytes, APTR buffer, LONG *res)
{
	if (fl->file == NULL)
		return ERROR_OBJECT_WRONG_TYPE;

	if ((*res = hfs_write(fl->file, buffer, bytes)) >= 0)
	{
		hfs_flush(sb->vol);
		return 0;
	}
	return ERROR_SEEK_ERROR;
}

LONG FileSeek(struct FSSuper *sb, struct ExtFileLock *fl, LONG offset, LONG mode, LONG *res)
{
	int from;
	LONG new_pos, old_pos;

	if (fl->file == NULL)
		return ERROR_OBJECT_WRONG_TYPE;

	if (mode == OFFSET_BEGINNING && offset >= 0)
		from = HFS_SEEK_SET;
	else if (mode == OFFSET_CURRENT)
		from = HFS_SEEK_CUR;
	else if (mode == OFFSET_END && offset <= 0)
		from = HFS_SEEK_END;
	else
		return ERROR_OBJECT_WRONG_TYPE;

	if ((old_pos = hfs_seek(fl->file, 0, HFS_SEEK_CUR)) >= 0)
	{
		if ((new_pos = hfs_seek(fl->file, offset, from)) >= 0)
		{
			*res = old_pos;
			fl->lastpos = new_pos;
			return 0;
		}
	}
	return ERROR_SEEK_ERROR;
}
 
LONG FileSetSize(struct FSSuper *sb, struct ExtFileLock *fl, LONG offset, LONG mode, LONG *res)
{
	LONG size, old_pos;
	ULONG newsize;

	if (fl->file == NULL)
		return ERROR_OBJECT_WRONG_TYPE;

	if ((old_pos = hfs_seek(fl->file, 0, HFS_SEEK_CUR)) >= 0)
	{
		if ((size = hfs_seek(fl->file, 0, HFS_SEEK_END)) >= 0)
		{
			if (mode == OFFSET_BEGINNING && offset >= 0)
				newsize	= offset;
			else if (mode == OFFSET_CURRENT && old_pos + offset >= 0)
				newsize = old_pos + offset;
			else if (mode == OFFSET_END && offset + size >= 0)
				newsize = offset + size;
			else
				return ERROR_OBJECT_WRONG_TYPE;

			if (hfs_truncate(fl->file, newsize) == 0)
			{
				hfs_flush(sb->vol);
				if (old_pos > newsize)
					if ((hfs_seek(fl->file, old_pos, HFS_SEEK_SET)) > 0)
					{
						*res = newsize;
						return 0;
					}
			}
		}
	}
	return ERROR_SEEK_ERROR;
}

/* disk info */

#ifndef ID_BUSY
#define ID_BUSY 0x42555359
#endif

void FillDiskInfo (struct InfoData *id)
{
	struct DosEnvec *de = BADDR(glob->fssm->fssm_Environ);

	memset(id, 0, sizeof(struct InfoData));

	id->id_NumSoftErrors = 0;
	id->id_UnitNumber = glob->fssm->fssm_Unit;

	if (glob->sb)
	{
		hfsvolent volent;

		if (hfs_vstat(glob->sb->vol, &volent) >= 0)
		{
			glob->sb->total_blocks = volent.totbytes >> HFS_BLOCKSZ_BITS;
			glob->sb->free_blocks = volent.freebytes >> HFS_BLOCKSZ_BITS;
		}

		id->id_NumBlocks = glob->sb->total_blocks;
		id->id_NumBlocksUsed = glob->sb->total_blocks - glob->sb->free_blocks;
		id->id_BytesPerBlock = glob->sb->blocksize;
		id->id_DiskState = (glob->sb->read_only) ? ID_WRITE_PROTECTED : ID_VALIDATED;
		id->id_DiskType = ID_HFS_DISK;
		id->id_VolumeNode = MKBADDR(glob->sb->doslist);
		id->id_InUse = (glob->sb->doslist->dol_misc.dol_volume.dol_LockList) ? DOSTRUE : DOSFALSE;
	}
	else
	{
		id->id_NumBlocks = de->de_Surfaces * de->de_BlocksPerTrack * (de->de_HighCyl+1 - de->de_LowCyl) / de->de_SectorPerBlock;
		id->id_NumBlocksUsed = id->id_NumBlocks;
		id->id_BytesPerBlock = de->de_SizeBlock << 2;
		id->id_DiskState = ID_WRITE_PROTECTED;
		if (glob->disk_inhibited)
			id->id_DiskType = ID_BUSY;
		else if (glob->disk_inserted)
			id->id_DiskType = ID_NOT_REALLY_DOS;
		else
			id->id_DiskType = ID_NO_DISK_PRESENT;
		id->id_VolumeNode = NULL;
		id->id_InUse = DOSFALSE;
	}
}

LONG RenameDisk(struct FSSuper *sb, UBYTE *newname, LONG newname_len)
{
	UBYTE oldname_arg[32];
	UBYTE newname_arg[32];

	if (newname_len > HFS_MAX_VLEN)
		return ERROR_LINE_TOO_LONG;
 
	memcpy(oldname_arg, &sb->name[1], sb->name[0]);
	oldname_arg[sb->name[0]] = ':';
	oldname_arg[sb->name[0]+1] = '\0';

	memcpy(newname_arg, newname, newname_len);
	newname_arg[newname_len] = ':';
	newname_arg[newname_len+1] = '\0';

	kprintf("\tRename from %s to %s\n", (LONG)oldname_arg, (LONG)newname_arg);
	if (hfs_rename(sb->vol, oldname_arg, newname_arg) == 0)
	{
		hfs_flush(sb->vol);
		memcpy(&sb->name[1], newname, newname_len);
		sb->name[newname_len+1] = '\0';
		sb->name[0] = newname_len;
 
		kprintf("\tNew name: %s\n", (ULONG)&sb->name[1]);
		return 0;
	}

	return ERROR_SEEK_ERROR;
}

LONG FormatDisk(UBYTE *name, LONG namelen)
{
	UBYTE volname[32];
	memcpy(volname, name, namelen);
	volname[namelen] = '\0';
 
	kprintf("\tFormating disk %s\n", (LONG) volname);

	if (hfs_format((const char *)glob->fssm, 0, HFS_MODE_ANY, volname, 0, NULL) == 0)
	{
		kprintf("\tDisk sucessfully formated\n");
		return 0;
	}
	else
	{
		kprintf("\tError formating disk: %s\n", (LONG)hfs_error);
	}
	return ERROR_SEEK_ERROR;
}

LONG SetDate(struct FSSuper *sb, struct ExtFileLock *fl, STRPTR path, LONG pathlen, struct DateStamp *datestamp)
{
	LONG err;
	ULONG cid, parid;

	kprintf("\tLooking for a target object, pathlen %ld\n", pathlen);

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &cid, &parid)) == 0)
	{
		hfsdirent ent;
		hfsdir *dir = NULL;

		if (parid == HFS_CNID_ROOTPAR)
			dir = hfs_opendir(sb->vol, "");
		else
		{
			if (hfs_setcwd(sb->vol, parid) == 0)
				dir = hfs_opendir(sb->vol, ":");
		}

		if (dir == NULL)
		{
			kprintf("\terror - cannot open parent directory for cid %d\n", cid);
			return ERROR_OBJECT_NOT_FOUND;
		}

		while (hfs_readdir(dir, &ent) == 0)
			if (ent.cnid == cid)
				break;

		hfs_closedir(dir);
		if (ent.cnid != cid)
			return ERROR_OBJECT_NOT_FOUND;

		/* hfsdirent acquired, change it and update disk structures */

		DateStamp2MacTime(datestamp, &ent.mddate);

		if (err == 0)
		{
			if (hfs_setattr(sb->vol, ent.name, &ent) == 0)
				hfs_flush(sb->vol);
			else
				err = ERROR_SEEK_ERROR;
		}
	}

	return err;
}

static char *blessed_creator = "UNIX";
static char *blessed_type = "tbxi";

LONG SetMacBootable(struct FSSuper *sb, struct ExtFileLock *fl)
{
	LONG cid = fl->fl_Key;
	LONG parid = fl->parid;
	hfsvolent volent;
	hfsdirent ent;
	hfsdir *dir = NULL;

	if (parid == HFS_CNID_ROOTPAR)
		dir = hfs_opendir(sb->vol, "");
	else
	{
		if (hfs_setcwd(sb->vol, parid) == 0)
			dir = hfs_opendir(sb->vol, ":");
	}

	if (dir == NULL)
	{
		kprintf("\terror - cannot open parent directory for cid %d\n", cid);
		return ERROR_OBJECT_NOT_FOUND;
	}

	while (hfs_readdir(dir, &ent) == 0)
		if (ent.cnid == cid)
			break;

	hfs_closedir(dir);
	if (ent.cnid != cid)
		return ERROR_OBJECT_NOT_FOUND;

	if ((ent.flags & HFS_ISDIR)) /* you cannot set mac bootable object to a directory */
		return ERROR_OBJECT_WRONG_TYPE;

	kprintf("\tFile %s, cnid %ld, parent %ld, creator %s type %s\n", (ULONG)ent.name, ent.cnid, parid, (ULONG)ent.u.file.creator, (ULONG)ent.u.file.type);

	memcpy(ent.u.file.creator, blessed_creator, 5);
	kprintf("\tChanged creator to UNIX\n");

	memcpy(ent.u.file.type, blessed_type, 5);
	kprintf("\tChanged type to tbxi\n");

	if (hfs_setattr(sb->vol, ent.name, &ent) == 0)
	{
		if (hfs_vstat(sb->vol, &volent) >= 0)
		{
			kprintf("\tCurrent blessed cnid: %ld\n", volent.blessed);

			kprintf("\tChanged blessed cnid to %ld\n", parid);
			volent.blessed = parid;

			if (hfs_vsetattr(sb->vol, &volent) == 0)
			{
				/* unbless all other files in the directory */

				kprintf("\tBlessed directory set\n");
				if (hfs_setcwd(sb->vol, parid) == 0)
					dir = hfs_opendir(sb->vol, ":");
 
				if (dir == NULL)
				{
					kprintf("\tFailed to open directory\n");
					goto skip;
				}

				kprintf("\tScanning for old blessed files\n");
				while (hfs_readdir(dir, &ent) == 0)
				{
					if (ent.cnid != cid && memcmp(ent.u.file.type, blessed_type, 5) == 0 && memcmp(ent.u.file.creator, blessed_creator, 5) == 0)
					{
						memcpy(ent.u.file.creator, RAW_CREA, 5);
						memcpy(ent.u.file.type, RAW_TYPE, 5);
						hfs_setattr(sb->vol, ent.name, &ent);
						kprintf("\tChanged type/creator for %s file to UNIX/????\n", ent.name);
					}
				}
				hfs_closedir(dir);
skip:
				/* flush changes to disk */
				hfs_flush(sb->vol);
				return 0;
			}
			kprintf("\thfsvsetattr failed\n");
		}
		kprintf("\thfsvstat failed\n");
	}
	kprintf("\thfs_setattr failed\n");
	return ERROR_SEEK_ERROR;
}

#if 0
/* Extended Attributes */

static LONG ExtAttr_GetControl(struct FSSuper *sb, hfsdirent *ent, struct ExtAttr *attrib)
{
	if (attrib->exa_Type != EXA_RAW && attrib->exa_Type != EXA_STRING)
		return ERROR_OBJECT_WRONG_TYPE;

	if (attrib->exa_Size < sizeof (ent->u.file.creator) && attrib->exa_Buffer != NULL)
		return ERROR_LINE_TOO_LONG;

	kprintf("\tFound %s attribute: '%s'\n", (LONG)attrib->exa_Name, (LONG)ent->u.file.creator);
	
	if (attrib->exa_Buffer)
		memcpy(attrib->exa_Buffer, ent->u.file.creator, sizeof(ent->u.file.creator));
	attrib->exa_Size = sizeof (ent->u.file.creator);

	return 0;
}

static LONG ExtAttr_SetControl(struct FSSuper *sb, hfsdirent *ent, struct ExtAttr *attrib)
{
	if (attrib->exa_Type != EXA_RAW && attrib->exa_Type != EXA_STRING)
		return ERROR_OBJECT_WRONG_TYPE;

	if (attrib->exa_Size >= 4 && attrib->exa_Buffer != NULL)
	{
		kprintf("\tChanging %s attribute to '%s'\n", (LONG)attrib->exa_Name, (LONG)attrib->exa_Buffer);
		memcpy(ent->u.file.creator, attrib->exa_Buffer, 4);
	}
	else
		return ERROR_OBJECT_WRONG_TYPE;

	return 0;
}

static LONG ExtAttr_GetType(struct FSSuper *sb, hfsdirent *ent, struct ExtAttr *attrib)
{
	if (attrib->exa_Type != EXA_RAW && attrib->exa_Type != EXA_STRING)
		return ERROR_OBJECT_WRONG_TYPE;

	if (attrib->exa_Size < sizeof (ent->u.file.type) && attrib->exa_Buffer != NULL)
		return ERROR_LINE_TOO_LONG;

	kprintf("\tFound %s attribute: '%s'\n", (LONG)attrib->exa_Name, (LONG)ent->u.file.type);
	
	if (attrib->exa_Buffer)
		memcpy(attrib->exa_Buffer, ent->u.file.type, sizeof(ent->u.file.type));
	attrib->exa_Size = sizeof (ent->u.file.type);

	return 0;
}

static LONG ExtAttr_SetType(struct FSSuper *sb, hfsdirent *ent, struct ExtAttr *attrib)
{
	if (attrib->exa_Type != EXA_RAW && attrib->exa_Type != EXA_STRING)
		return ERROR_OBJECT_WRONG_TYPE;

	if (attrib->exa_Size >= 4 && attrib->exa_Buffer != NULL)
	{
		kprintf("\tChanging %s attribute to '%s'\n", (LONG)attrib->exa_Name, (LONG)attrib->exa_Buffer);
		memcpy(ent->u.file.type, attrib->exa_Buffer, 4);
	}
	else
		return ERROR_OBJECT_WRONG_TYPE;

	return 0;
}
 
static LONG ExtAttr_GetBlessed(struct FSSuper *sb, hfsdirent *ent, struct ExtAttr *attrib)
{
	if (attrib->exa_Type != EXA_RAW && attrib->exa_Type != EXA_BOOLEAN)
		return ERROR_OBJECT_WRONG_TYPE;

	if (attrib->exa_Size < sizeof(LONG) && attrib->exa_Buffer != NULL)
		return ERROR_LINE_TOO_LONG;

	kprintf("\tFound %s attribute: %ld\n", (LONG)attrib->exa_Name, (ent->cnid == sb->blessed_cid));
	if (attrib->exa_Buffer)
		*((LONG *)attrib->exa_Buffer) = (ent->cnid == sb->blessed_cid);
	attrib->exa_Size = sizeof (LONG);

	return 0;
}

static LONG ExtAttr_SetBlessed(struct FSSuper *sb, hfsdirent *ent, struct ExtAttr *attrib)
{
	BOOL blessed;
	hfsvolent volent;
 
	if (attrib->exa_Type != EXA_RAW && attrib->exa_Type != EXA_BOOLEAN)
		return ERROR_OBJECT_WRONG_TYPE;

	if (attrib->exa_Size != 4 || attrib->exa_Buffer == NULL)
		return ERROR_OBJECT_WRONG_TYPE;

	blessed = *((LONG *)attrib->exa_Buffer);

	if (hfs_vstat(sb->vol, &volent) >= 0)
	{
		kprintf("\tChanging %s attribute: %ld\n", (LONG)attrib->exa_Name, blessed);
		if (blessed)
		{
			volent.blessed = ent->cnid;
			sb->blessed_cid = ent->cnid;
		}
		else if (volent.blessed == ent->cnid)
		{
			volent.blessed = 0;
			sb->blessed_cid = 0;
		}

		if (hfs_vsetattr(sb->vol, &volent) != 0)
			return ERROR_SEEK_ERROR;
	}
	return 0;
}
 
static struct ExtAttrEntry
{
	STRPTR name;
	LONG (*func_get)(struct FSSuper *sb, hfsdirent *entry, struct ExtAttr *attr);
	LONG (*func_set)(struct FSSuper *sb, hfsdirent *entry, struct ExtAttr *attr);
} ExtAttrArray[] = {
	{ "HFS_CONTROL", ExtAttr_GetControl, ExtAttr_SetControl },
	{ "HFS_TYPE", ExtAttr_GetType, ExtAttr_SetType },
	{ "HFS_BLESSED", ExtAttr_GetBlessed, ExtAttr_SetBlessed },
	{ NULL, NULL, NULL },
};

LONG GetExtAttr(struct FSSuper *sb, struct ExtFileLock *fl, STRPTR path, LONG pathlen, struct ExtAttr *attrib)
{
	LONG err;
	ULONG cid, parid;

	kprintf("\tLooking for a target object, pathlen %ld\n", pathlen);

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &cid, &parid)) == 0)
	{
		hfsdirent ent;
		hfsdir *dir = NULL;

		if (parid == HFS_CNID_ROOTPAR)
			dir = hfs_opendir(sb->vol, "");
		else
		{
			if (hfs_setcwd(sb->vol, parid) == 0)
				dir = hfs_opendir(sb->vol, ":");
		}

		if (dir == NULL)
		{
			kprintf("\terror - cannot open parent directory for cid %d\n", cid);
			return ERROR_OBJECT_NOT_FOUND;
		}

		while (hfs_readdir(dir, &ent) == 0)
			if (ent.cnid == cid)
				break;

		hfs_closedir(dir);
		if (ent.cnid != cid)
			return ERROR_OBJECT_NOT_FOUND;

		/* hfsdirent acquired, lets scan attributes and fill data... */

		if (! (ent.flags & HFS_ISDIR)) /* scan only files ... */
		{
			while (attrib)
			{
				struct ExtAttrEntry *exap = ExtAttrArray;
				err = ERROR_INVALID_COMPONENT_NAME;

				if (attrib->exa_Name == NULL)
					break;

				kprintf("\tLooking for %s attribute\n", (LONG)attrib->exa_Name);

				while (exap->name != NULL)
				{
					if (Stricmp(attrib->exa_Name, exap->name) == 0)
					{
						if ((err = exap->func_get(sb, &ent, attrib)) == 0)
							break;
					}
					exap += 1;
				}                                         

				if (err != 0)
				{
					/* err != 0 - no such attribute (ERROR_INVALID_COMPONENT_NAME or attribute read error */
					break;
				}
				attrib = attrib->exa_Next;
			}
		}
		else
			err = ERROR_INVALID_COMPONENT_NAME; /* no attributes for directories are supported yet ... */
	}

	return err;
}

LONG SetExtAttr(struct FSSuper *sb, struct ExtFileLock *fl, STRPTR path, LONG pathlen, struct ExtAttr *attrib)
{
	LONG err;
	ULONG cid, parid;

	kprintf("\tLooking for a target object, pathlen %ld\n", pathlen);

	if ((err = FindObjectByPath(sb, fl, path, pathlen, &cid, &parid)) == 0)
	{
		hfsdirent ent;
		hfsdir *dir = NULL;

		if (parid == HFS_CNID_ROOTPAR)
			dir = hfs_opendir(sb->vol, "");
		else
		{
			if (hfs_setcwd(sb->vol, parid) == 0)
				dir = hfs_opendir(sb->vol, ":");
		}

		if (dir == NULL)
		{
			kprintf("\terror - cannot open parent directory for cid %d\n", cid);
			return ERROR_OBJECT_NOT_FOUND;
		}

		while (hfs_readdir(dir, &ent) == 0)
			if (ent.cnid == cid)
				break;

		hfs_closedir(dir);
		if (ent.cnid != cid)
			return ERROR_OBJECT_NOT_FOUND;

		/* hfsdirent acquired, lets scan attributes and fill data... */

		if (! (ent.flags & HFS_ISDIR)) /* scan only files ... */
		{
			while (attrib)
			{
				struct ExtAttrEntry *exap = ExtAttrArray;
				err = ERROR_INVALID_COMPONENT_NAME;

				if (attrib->exa_Name == NULL)
					break;

				kprintf("\tLooking for %s attribute\n", (LONG)attrib->exa_Name);

				while (exap->name != NULL)
				{
					if (Stricmp(attrib->exa_Name, exap->name) == 0)
					{
						if ((err = exap->func_set(sb, &ent, attrib)) == 0)
							break;
					}
					exap += 1;
				}                                         

				if (err != 0)
				{
					/* err != 0 - no such attribute (ERROR_INVALID_COMPONENT_NAME or attribute read error */
					break;
				}
				attrib = attrib->exa_Next;
			}
		}
		else
			err = ERROR_INVALID_COMPONENT_NAME; /* no attributes for directories are supported yet ... */

		if (err == 0)
		{
			if (hfs_setattr(sb->vol, ent.name, &ent) == 0)
				hfs_flush(sb->vol);
			else
				err = ERROR_SEEK_ERROR;
		}
	}

	return err;
}
 
#endif

